/**
 * LightPreview is a very lightweight code previewer designed to load your code sample
 *   inside of an iFrame for encapsulated viewing. `<light-preview>` uses PrismJS as it's
 * @customElement
 * @tagname light-preview
 *
 * @csspart base - The base wrapping element
 * @csspart preview - The wrapper around the iframe / div preview showing your code being run in the browser.
 * @csspart start-panel - The iframe or preview-div
 * @csspart preview-div - used for inline previews
 * @csspart iframe - used for iframe based previews (default)
 * @csspart panel-resizer - The button that resizes the panel
 * @csspart end-panel - The white space when dragging the resizer
 * @csspart source-details - The details element that shows the source code
 * @csspart code-wrapper - The div that wraps the <pre> + <code> elements containing your highlighted code
 * @csspart pre - The <pre> element wrapping the source code
 * @csspart code - The <code>  element wrapping the source code
 * @csspart source-code-toggle - The button to show the source code
 * @csspart actions - The buttons at the bottom of the code previewer
 * @csspart before-expanded-code - A wrapper around the "before-expanded-code" slot
 * @csspart after-expanded-code - A wrapper around the "after-expanded-code" slot
 * @csspart after-actions - A wrapper around the "after-actions" slot
 * @csspart source-code-toggle-icon - The caret icon in the source code toggle
 *
 * @slot resize-icon - The icon to display in the resizer button. Override this to provide your own icon.
 * @slot summary - What to display in source code expander
 * @slot source-code-toggle-icon - Slot in your own icon to override the default caret.
 * @slot above-actions - Additional actions placed above the source code toggle
 * @slot below-actions - Additional actions placed above the source code toggle
 * @slot actions - Slot in buttons / links to allow for additional actions in the bottom bar. These are placed inline.
 * @slot preview-html - If you want to run code that is slightly different from the source code you want to display, slot it into "preview-html"
 * @slot code - Used to display both source code and to power your preview in the iframe. If you slot in "preview-html", then it will only be used to show / highlight your source code.
 * @slot before-expanded-code - A slot to add actions such as buttons. This will sit above the source-code-toggle, and when code is expanded, will be *above* the expanded code.
 * @slot after-expanded-code - A slot to add actions such as buttons. This will sit above the source-code-toggle, and when code is expanded, will be *below* the expanded code.
 * @slot after-actions - A slot add actions below the source-code-toggle button.
 */
export default class LightPreview extends BaseElement {
    /**
     * @override
     */
    static override dependencies: {
        "light-disclosure": typeof LightDisclosure;
        "light-code": typeof LightCode;
    };
    /**
     * @override
     */
    static override styles: import("lit").CSSResult[];
    /**
     * @override
     */
    static override properties: {
        summary: {};
        sandboxSettings: {
            reflect: boolean;
            attribute: string;
        };
        previewMode: {
            reflect: boolean;
            attribute: string;
        };
        open: {
            reflect: boolean;
            type: BooleanConstructor;
        };
        resizePosition: {
            reflect: boolean;
            type: NumberConstructor;
            attribute: string;
        };
        resizing: {
            reflect: boolean;
            type: BooleanConstructor;
        };
        scriptScope: {
            attribute: string;
        };
        previewHtml: {
            attribute: string;
        };
        unescapeBehavior: {
            attribute: string;
        };
        disableHighlight: {
            type: BooleanConstructor;
            attribute: string;
        };
        preserveWhitespace: {
            type: BooleanConstructor;
            attribute: string;
        };
        highlightLines: {
            attribute: string;
        };
        insertedLines: {
            attribute: string;
        };
        deletedLines: {
            attribute: string;
        };
        disableLineNumbers: {
            type: BooleanConstructor;
            reflect: boolean;
            attribute: string;
        };
        lineNumberStart: {
            type: NumberConstructor;
            attribute: string;
        };
        wrap: {
            reflect: boolean;
            attribute: string;
        };
        language: {};
        code: {};
        highlighter: {
            attribute: boolean;
            state: boolean;
        };
    };
    /**
     * The sandbox settings to provide to the <iframe>
     * @type {string}
     */
    sandboxSettings: string;
    /**
     * The text to provide in the <details> toggle button
     * @type {string}
     */
    summary: string;
    /**
     * Set to true to not use an <iframe> for previewing
     * @type {"iframe" | "shadow-dom"}
     */
    previewMode: "iframe" | "shadow-dom";
    /**
     * When the resizer is being dragged, this will be true.
     * @type {boolean}
     */
    resizing: boolean;
    /**
     * If `disableHighlight` is true, then you must pass in an element into `previewHtml` to be able to get
     *   the code to run in the previewer.
     * @type {string}
     */
    previewHtml: string;
    /**
     * Whether or not the source code is being shown
     * @type {boolean}
     */
    open: boolean;
    /**
     * The current position of the resizer. 100 means all the way to right. 0 means all the way to left.
     * @type {number}
     */
    resizePosition: number;
    /**
     * @internal
     * @type {ResizeObserver}
     */
    resizeObserver: ResizeObserver;
    /**
     * @internal
     * @type {MutationObserverInit}
     */
    __mutationObserverConfig: MutationObserverInit;
    /**
     * @internal
     * @type {() => void}
     */
    previewHtmlDebounce: () => void;
    /**
     * @internal
     * @type {() => void}
     */
    codeDebounce: () => void;
    /**
     * When using `preview-mode="shadow-dom"`,
     * There's a funky issue with previews where if you want the location of the shadowRoot
     * you are, you can't get it. As a result, `<light-preview>` supports the idea of a "scriptScope"
     * where `document` is bound to the current shadowRoot instead of the actual top level `document`
     * For more info, check out this GitHub issue:
     * @link {https://github.com/WICG/webcomponents/issues/717#issuecomment-1126786185}
     *
     * @type {"document" | "shadow-dom"}
     */
    scriptScope: "document" | "shadow-dom";
    /**
     * The language to highlight for.
     * @type {string}
     */
    language: string;
    /**
     * Preserve leading and trailing whitespace.
     */
    preserveWhitespace: boolean;
    /**
     * If disabled, its on you to provide `<pre><code></code></pre>`
     * @type {boolean}
     */
    disableHighlight: boolean;
    /**
     * @type {string} - A string of possible lines to highlight. Example: "{1-9, 16, 18}"
     */
    highlightLines: string;
    /**
     * @type {string} - A string of lines that are inserted for diffs. Example: "{1-9, 16, 18}"
     */
    insertedLines: string;
    /**
     * @type {string} - A string of lines that are deleted for diffs. Example: "{1-9, 16, 18}"
     */
    deletedLines: string;
    /**
     * @type {boolean} whether or not to disable line numbers
     */
    disableLineNumbers: boolean;
    /**
     * Where to start counting indexes in the gutter. Note, this is purely for display purposes.
     * @type {number}
     */
    lineNumberStart: number;
    /**
     * We will take the code, wrap it in `<pre><code></code></pre>` and run it through
     * Prism.js.
     * If the element has `disableHighlight`, we will not touch their code. Instead they must pass in escapedHTML.
     * @type {string}
     */
    code: string;
    /**
     * Points to an instance of Prism from "prism-esm" for adjusting highlighting, adding plugins, etc.
     * @type {ReturnType<typeof createPrismInstance>}
     */
    highlighter: ReturnType<typeof createPrismInstance>;
    /**
     * @property
     * @type {"soft" | "hard"}
     * If `wrap="soft"`, lines will wrap when they reach the edge of their container. If `wrap="none"`, lines will not wrap instead all the user to scroll horizontally to see more code.
     */
    wrap: "soft" | "hard";
    /**
     * Whether or not to transform `&lt;/script>` into `<script>`
     * If true, will run transform. If false, will leave the code as is.
     *
     * @type {"all" | "last" | "none"}
     */
    unescapeBehavior: "all" | "last" | "none";
    /**
     * @internal
     * Reinstalls the mutation on slotted preview-html
     */
    resetIframeCodeMutationObserver(): void;
    previewHtmlMutationObserver: MutationObserver | undefined;
    /**
     * @internal
     * Reinstalls the mutation observer on slotted code
     */
    resetCodeMutationObserver(): void;
    codeMutationObserver: MutationObserver | undefined;
    /**
     * @internal
     * @param {"preview-html" | "code"} variable
     */
    handleMutation(variable: "preview-html" | "code"): void;
    /**
     * @internal
     * @param {string} name
     * @returns {HTMLSlotElement | null | undefined}
     */
    findSlot(name: string): HTMLSlotElement | null | undefined;
    /**
     * @internal
     * @param {Event | { target?: undefined | null | HTMLSlotElement }} e
     */
    handleTemplate(e: Event | {
        target?: undefined | null | HTMLSlotElement;
    }): void;
    /**
     * @internal
     */
    updateIframeContent(): void;
    /**
     * @internal
     * @param {string} text
     */
    escapeCharacters(text: string): string;
    /**
     * Only used to unescape things like `&lt;/script>` from slotted content.
     * @internal
     * @param {string} text
     */
    transformTags(text: string): string;
    /**
     * @override
     * @internal
     * @param {import("lit").PropertyValues<this>} changedProperties
     */
    override willUpdate(changedProperties: import("lit").PropertyValues<this>): void;
    _iframeDebounce: NodeJS.Timeout | undefined;
    /**
     * @internal
     * Sets an initial width so we dont need to keep computing getBoundingClientRect
     */
    updateCachedWidth(): void;
    cachedWidth: number | undefined;
    /**
     * @internal
     * @param {ResizeObserverEntry[]} entries
     */
    handleResize(entries: ResizeObserverEntry[]): void;
    /**
     * @public
     * Override this function to use your own highlighter
     */
    public highlight(code?: string): string;
    /**
     * @private
     * @param {HTMLElement | Element | ShadowRoot} root
     */
    private addShadowRootToPreview;
    /**
     * @override
     */
    override render(): import("lit").TemplateResult<1>;
    /**
     * @internal
     */
    get panelResizer(): Element | null | undefined;
    /** Drag stuff */
    /**
     * @internal
     * @param {PointerEvent} event
     */
    handleDrag(event: PointerEvent): void;
    /**
     * @internal
     */
    updateResizePosition(resizePosition?: number): void;
    /**
     * @internal
     * @param {KeyboardEvent} event
     */
    handleResizerKeydown(event: KeyboardEvent): void;
    /**
     * @internal
     * @param {number} pixels
     * @return {number}
     */
    pixelsToPercentage(pixels: number): number;
}
import { BaseElement } from "../../../internal/base-element.js";
import { createPrismInstance } from "../../../internal/prism-highlight.js";
import LightDisclosure from "../light-disclosure/light-disclosure.js";
import LightCode from "../light-code/light-code.js";
